"""
Created on Wed April 8 20:30:00 2024

@author: Anna Grim
@email: anna.grim@alleninstitute.org



"""

import numpy as np

from segmentation_skeleton_metrics import utils


def detect_potentials(labeled_graphs, get_labels):
    """
    Detects merges between ground truth graphs which are considered to be
    potential merge sites.

    Parameters
    ----------
    labeled_graphs : dict
        Dictionary where the keys are graph ids and the values are the
        corresponding graphs.
    get_labels : func
        Gets the label of a node in "labeled_graphs".

    Returns
    -------
    merge_ids : set[tuple]
        Set of tuples containing a tuple of graph ids and common label between
        the graphs.

    """
    merge_ids = set()
    visited = set()
    for id_1 in labeled_graphs.keys():
        for id_2 in labeled_graphs.keys():
            ids = frozenset((id_1, id_2))
            if id_1 != id_2 and ids not in visited:
                visited.add(ids)
                intersection = get_labels(id_1).intersection(get_labels(id_2))
                for label in intersection:
                    merge_ids.add((ids, label))
    return merge_ids


def localize(graph_1, graph_2, merged_1, merged_2, dist_threshold, merge_id):
    """
    Finds the closest pair of xyz coordinates from "merged_1" and "merged_2".

    Parameters
    ----------
    graph_1 : networkx.Graph
        Graph with potential merge.
    graph_2 : networkx.Graph
        Graph with potential merge.
    merged_1 : set
        Nodes contained in "graph_1" with same labels as nodes in "merged_2".
    merged_2 : set
        Nodes contained in "graph_2" with same labels as nodes in "merged_1".
    dist_threshold : float
        Distance that determines whether two graphs contain a merge site.
    merge_id : tuple
        Tuple containing ids corresponding to "graph_1" and "graph_2" along
        the common label between them.

    Returns
    -------
    xyz_pair : list[numpy.ndarray]
        Closest pair of xyz coordinates from "merged_1" and "merged_2".
    min_dist : float
        Distance between xyz coordinates in "xyz_pair".

    """
    # Check whether merge is spurious
    if len(merged_1) < 16 and len(merged_2) < 16:
        return merge_id, None, np.inf

    # Compute pairwise distances
    min_dist = np.inf
    xyz_pair = list()
    for i in merged_1:
        for j in merged_2:
            xyz_i = graph_1.nodes[i]["xyz"]
            xyz_j = graph_2.nodes[j]["xyz"]
            if utils.dist(xyz_i, xyz_j) < min_dist:
                min_dist = utils.dist(xyz_i, xyz_j)
                xyz_pair = [xyz_i, xyz_j]
                if min_dist < dist_threshold:
                    return merge_id, xyz_pair, min_dist
    return merge_id, xyz_pair, min_dist


def intersections(get_projection, xyz_list, xyz_to_id, close_dist_threshold):
    """
    Projects coordinates in "xyz_list" onto ground truth graphs, then stores
    the correspond label in "hit_ids" if the projection distance is less than
    "dist_threshold".

    Parameters
    ----------
    get_projection : func
        Computes projections onto ground truth graphs.
    xyz_list : list
        List of xyz coordinates to be projected.
    xyz_to_id : dict
        Dictionary where keys are xyz coordinates in the ground truth graphs
        and values are a dictionary that stores the corresonding label and
        node ids.
    close_dist_threshold : float
        Threshold that defines whether a projection distance is small enough
        to be considered a valid intersection.

    Returns
    -------
    hit_ids : dict
        Dictionary where deys are labels and values are node ids.

    """
    hit_ids = dict()
    multi_hits = set()
    for xyz in xyz_list:
        hat_xyz, d = get_projection(xyz)
        if d < close_dist_threshold:
            hits = list(xyz_to_id[hat_xyz].keys())
            if len(hits) > 1:
                multi_hits.add(hat_xyz)
            else:
                hat_i = xyz_to_id[hat_xyz][hits[0]]
                hit_ids = utils.append_dict_value(hit_ids, hits[0], hat_i)
    return utils.resolve(multi_hits, hit_ids, xyz_to_id)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

import os
from io import BytesIO
from time import time
from zipfile import ZipFile

import networkx as nx
import numpy as np
import tensorstore as ts
from scipy.spatial import distance

SUPPORTED_DRIVERS = ["neuroglancer_precomputed", "n5", "zarr"]


# -- os utils ---
def listdir(directory, ext=None):
    """
    Lists all files in "directory". If an extension "ext" is
    provided, then only files containing "ext" are returned.

    Parameters
    ----------
    directory : str
        Path to directory to be searched.

    ext : str, optional
       Extension of file type of interest. The default is None.

    Returns
    -------
    list
        List of all files in directory at "path" with extension "ext" if
        provided. Otherwise, list of all files in directory.

    """
    if ext is None:
        return [f for f in os.listdir(directory)]
    else:
        return [f for f in os.listdir(directory) if ext in f]


def list_paths(directory, ext=None):
    """
    Lists all paths within "directory".

    Parameters
    ----------
    directory : str
        Directory to be searched.
    ext : str, optional
        If provided, only paths of files with the extension "ext" are
        returned. The default is None.

    Returns
    -------
    list[str]
        List of all paths within "directory".

    """
    paths = []
    for f in listdir(directory, ext=ext):
        paths.append(os.path.join(directory, f))
    return paths


# --- io utils ---
def open_tensorstore(path, driver):
    """
    Uploads segmentation mask stored as a directory of shard files.

    Parameters
    ----------
    path : str
        Path to directory containing shard files.
    driver : str
        Storage driver needed to read data at "path".

    Returns
    -------
    sparse_volume : dict
        Sparse image volume.

    """
    assert driver in SUPPORTED_DRIVERS, "Driver is not supported!"
    arr = ts.open(
        {
            "driver": driver,
            "kvstore": {
                "driver": "gcs",
                "bucket": "allen-nd-goog",
                "path": path,
            },
            "context": {
                "cache_pool": {"total_bytes_limit": 1000000000},
                "cache_pool#remote": {"total_bytes_limit": 1000000000},
                "data_copy_concurrency": {"limit": 8},
            },
            "recheck_cached_data": "open",
        }
    ).result()
    if driver == "neuroglancer_precomputed":
        return arr[ts.d["channel"][0]]
    return arr


def read_tensorstore(path):
    """
    Reads neuroglancer_precomputed file at "path".

    Parameters
    ----------
    path : str
        Path to directory containing shardsS.

    Returns
    -------
    ts.TensorStore
        Image volume.

    """
    dataset_ts = ts.open(
        {
            "driver": "neuroglancer_precomputed",
            "kvstore": {"driver": "file", "path": path},
        }
    ).result()
    return dataset_ts[ts.d["channel"][0]]


def read_from_gcs_zip(zip_file, path):
    """
    Reads the content of an swc file from a zip file in a GCS bucket.

    """
    with zip_file.open(path) as text_file:
        return text_file.read().decode("utf-8").splitlines()


def read_txt(path):
    """
    Reads txt file stored at "path".

    Parameters
    ----------
    path : str
        Path where txt file is stored.

    Returns
    -------
    str
        Contents of txt file.

    """
    with open(path, "r") as f:
        return f.readlines()


def list_gcs_filenames(bucket, cloud_path, extension):
    """
    Lists all files in a GCS bucket with the given extension.

    """
    blobs = bucket.list_blobs(prefix=cloud_path)
    return [blob.name for blob in blobs if extension in blob.name]


def list_files_in_gcs_zip(zip_content):
    """
    Lists all files in a zip file stored in a GCS bucket.

    """
    with ZipFile(BytesIO(zip_content), "r") as zip_file:
        return zip_file.namelist()


# -- miscellaneous --
def check_edge(edge_list, edge):
    """
    Checks if "edge" is in "edge_list".

    Parameters
    ----------
    edge_list : list or set
        List or set of edges.
    edge : tuple
        Edge.

    Returns
    -------
    bool : bool
        Indication of whether "edge" is contained in "edge_list".

    """
    if edge in edge_list or (edge[1], edge[0]) in edge_list:
        return True
    else:
        return False


def dist(v_1, v_2):
    """
    Computes distance between "v_1" and "v_2".

    Parameters
    ----------
    v_1 : np.ndarray
        Vector.
    v_2 : np.ndarray
        Vector.

    Returns
    -------
    float
        Distance between "v_1" and "v_2".

    """
    return distance.euclidean(v_1, v_2)


def get_midpoint(xyz_1, xyz_2):
    """
    Computes the midpoint between "xyz_1" and "xyz_2".

    Parameters
    ----------
    xyz_1 : numpy.ndarray
        xyz coordinate.
    xyz_2 : numpy.ndarray
        xyz coordinate.

    Returns
    -------
    numpy.ndarray
        Midpoint between "xyz_1" and "xyz_2".

    """
    return np.array([np.mean([xyz_1[i], xyz_2[i]]) for i in range(3)])


def to_world(xyz, anisotropy):
    """
    Converts "xyz" from image coordinates to real-world coordinates.

    Parameters
    ----------
    xyz : tuple or list
        Coordinates to be transformed.
    anisotropy : list[float]
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files.

    Returns
    -------
    list[float]
        Transformed coordinates.

    """
    return [xyz[i] * anisotropy[i] for i in range(3)]


def above_threshold(my_dict):
    # Find keys to delete
    delete_keys = list()
    for key, value in my_dict.items():
        if len(value) < 16:
            delete_keys.append(key)

    # Delete keys
    while len(delete_keys) > 0:
        key = delete_keys.pop()
        del my_dict[key]

    return my_dict


def resolve(multi_hits, dists, xyz_to_id_node):
    """
    Resolves discrepancy when xyz coordinates project onto ground truth graphs
    that contain nodes with same xyz coordinates.

    Parameters
    ----------
    multi_hits : set
        xyz coordinates that are common across multiple graphs.
    dists : dict
        Dictionary containing graph ids that predicted swc file has
        intersected.
    xyz_to_id_node : dict
        Dictionary where the keys are xyz coordinates that correspond to nodes
        in ground truth graphs and values are the graph ids and nodes.

    Return
    ------
    dists : dict
        Updated dictionary with discrepancies resolved.

    """
    for hat_xyz in multi_hits:
        keys = list(xyz_to_id_node[hat_xyz].keys())
        id = find_best(dists, keys)
        if id:
            node = xyz_to_id_node[hat_xyz][id]
            dists = append_dict_value(dists, id, node)
    return dists


def append_dict_value(my_dict, key, value):
    """
    Appends "value" to the list stored at "key".

    Parameters
    ----------
    my_dict : dict
        Dictionary to be queried.
    key : hashable data type
        Key to be query.
    value : list item type
        Value to append to list stored at "key".

    Returns
    -------
    my_dict : dict
        Updated dictionary.

    """
    if key in my_dict.keys():
        my_dict[key].append(value)
    else:
        my_dict[key] = [value]
    return my_dict


def find_best(my_dict, keys):
    best_key = None
    best_vote_cnt = 0
    if len(my_dict) > 0:
        for key in keys:
            vote_cnt = len(my_dict[key]) if key in my_dict.keys() else 0
            if vote_cnt > best_vote_cnt:
                best_key = key
                best_vote_cnt = vote_cnt
    return best_key


def get_id(path):
    """
    Gets segment id of the swc file at "path".

    Parameters
    ----------
    path : str
        Path to an swc file

    Return
    ------
    Segment id of swc file.

    """
    filename = os.path.basename(path)
    return filename.split(".")[0]


def equiv_class_mappings(connections_path, labels):
    label_to_class = {0: 0}
    labels_graph = build_labels_graph(connections_path, labels)
    for i, component in enumerate(nx.connected_components(labels_graph)):
        i += 1
        for label in component:
            label_to_class[label] = i
    return label_to_class


def build_labels_graph(connections_path, labels):
    labels_graph = nx.Graph()
    labels_graph.add_nodes_from(list(labels.keys()))
    for line in read_txt(connections_path):
        ids = line.split(",")
        id_1 = int(ids[0])
        id_2 = int(ids[1])
        labels_graph.add_edge(id_1, id_2)
    return labels_graph


# -- runtime --
def time_writer(t, unit="seconds"):
    """
    Converts runtime "t" to its proper unit.

    Parameters
    ----------
    t : float
        Runtime to be converted.
    unit : str, optional
        Unit of "t".

    Returns
    -------
    t : float
        Converted runtime.
    unit : str
        Unit of "t"

    """
    assert unit in ["seconds", "minutes", "hours"]
    upd_unit = {"seconds": "minutes", "minutes": "hours"}
    if t < 60 or unit == "hours":
        return t, unit
    else:
        t /= 60
        unit = upd_unit[unit]
        t, unit = time_writer(t, unit=unit)
    return t, unit


def init_timers():
    """
    Initializes two timers.

    Parameters
    ----------
    None

    Returns
    -------
    time.time
        Timer.
    time.time
        Timer.

    """
    return time(), time()


def progress_bar(current, total, bar_length=50):
    """
    Reports the progress of completing some process.

    Parameters
    ----------
    current : int
        Current iteration of process.
    total : int
        Total number of iterations to be completed
    bar_length : int, optional
        Length of progress bar

    Returns
    -------
    None

    """
    progress = int(current / total * bar_length)
    bar = (
        f"[{'=' * progress}{' ' * (bar_length - progress)}] {current}/{total}"
    )
    print(f"\r{bar}", end="", flush=True)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

from concurrent.futures import (
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    as_completed,
)
from time import time

import networkx as nx
import numpy as np
import tensorstore as ts
from scipy.spatial import KDTree

from segmentation_skeleton_metrics import graph_utils as gutils
from segmentation_skeleton_metrics import (
    merge_detection,
    split_detection,
    swc_utils,
    utils,
)
from segmentation_skeleton_metrics.swc_utils import save, to_graph

CLOSE_DIST_THRESHOLD = 3.5
INTERSECTION_THRESHOLD = 16
MERGE_DIST_THRESHOLD = 20


class SkeletonMetric:
    """
    Class that evaluates the quality of a predicted segmentation by comparing
    the ground truth skeletons to the predicted segmentation mask. The
    accuracy is then quantified by detecting splits and merges, then computing
    the following metrics:
        (1) Number of splits
        (2) Number of merges
        (3) Percentage of omit edges
        (4) Percentage of merged edges
        (5) Edge accuracy
        (6) Expected Run Length (ERL)

    """

    def __init__(
        self,
        pred_labels,
        pred_swc_paths,
        target_swc_paths,
        anisotropy=[1.0, 1.0, 1.0],
        black_holes_xyz_id=None,
        black_hole_radius=24,
        connections_path=None,
        ignore_boundary_mistakes=False,
        output_dir=None,
        valid_size_threshold=40,
        save_swc=False,
    ):
        """
        Constructs skeleton metric object that evaluates the quality of a
        predicted segmentation.

        Parameters
        ----------
        pred_labels : numpy.ndarray or tensorstore.TensorStore
            Predicted segmentation mask.
        target_swc_paths : list[str]
            List of paths to swc files where each file corresponds to a
            neuron in the ground truth.
        anisotropy : list[float], optional
            Image to real-world coordinates scaling factors applied to swc
            files. The default is [1.0, 1.0, 1.0]
        pred_swc_paths : list[str] or dict
            If swc files are on local machine, list of paths to swc files where
            each file corresponds to a neuron in the prediction. If swc files
            are on cloud, then dict with keys "bucket_name" and "path".
        black_holes_xyz_id : list, optional
            ...
        black_hole_radius : float, optional
            ...
        connections_path : list[tuple]
            Path to a txt file containing pairs of swc ids from the prediction
            that were predicted to be connected.
        ignore_boundary_mistakes : bool, optional
            Indication of whether to ignore mistakes near boundary of bounding
            box. The default is False.
        output_dir : str, optional
            Path to directory that each mistake site is written to. The default
            is None.
        valid_size_threshold : int, optional
            Threshold on the number of nodes contained in an swc file. Only swc
            files with more than "valid_size_threshold" nodes are stored in
            "self.valid_labels". The default is 40.
        save_swc : bool, optional
            Indication of whether to write mistake sites to an swc file. The
            default is False.

        Returns
        -------
        None.

        """
        # Store options
        self.anisotropy = anisotropy
        self.ignore_boundary_mistakes = ignore_boundary_mistakes
        self.output_dir = output_dir
        self.save = save_swc

        self.init_black_holes(black_holes_xyz_id)
        self.black_hole_radius = black_hole_radius

        # Labels
        self.label_mask = pred_labels
        self.valid_labels = None
        # swc_utils.parse(
        #      pred_swc_paths, valid_size_threshold, anisotropy
        # )
        self.init_equiv_labels(connections_path)

        # Build Graphs
        self.graphs = self.init_graphs(target_swc_paths, anisotropy)
        self.init_labeled_graphs()

        # Build kdtree
        self.init_xyz_to_id_node()
        self.init_kdtree()
        self.rm_spurious_intersections()

    # -- Initialize and Label Graphs --
    def init_equiv_labels(self, path):
        if path:
            assert self.valid_labels is not None, "Must provide valid labels!"
            self.equiv_labels_map = utils.equiv_class_mappings(
                path, self.valid_labels
            )
            valid_labels = dict()
            for label, values in self.valid_labels.items():
                equiv_label = self.equiv_labels_map[label]
                valid_labels[equiv_label] = values
            self.valid_labels = valid_labels
        else:
            self.equiv_labels_map = None

    def init_graphs(self, paths, anisotropy):
        """
        Initializes "self.graphs" by iterating over "paths" which
        correspond to neurons in the ground truth.

        Parameters
        ----------
        paths : list[str]
            List of paths to swc files which correspond to neurons in the
            ground truth.
        anisotropy : list[float]
            Image to real-world coordinates scaling factors applied to swc
            files.

        Returns
        -------
        None

        """
        graphs = dict()
        for path in paths:
            print(path)
            id = utils.get_id(path)
            graphs[id] = to_graph(path, anisotropy=anisotropy)
        return graphs

    def init_labeled_graphs(self):
        """
        Initializes "self.labeled_graphs" by copying each graph in
        "self.graphs", then labels each node with the label in
        "self.label_mask" that coincides with it.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        print("Labelling Graphs...")
        t0 = time()
        self.labeled_graphs = dict()
        self.id_to_label_nodes = dict()  # {id: {label: nodes}}
        for cnt, (id, graph) in enumerate(self.graphs.items()):
            utils.progress_bar(cnt + 1, len(self.graphs))
            labeled_graph, id_to_label_nodes = self.label_graph(graph)
            self.labeled_graphs[id] = labeled_graph
            self.id_to_label_nodes[id] = id_to_label_nodes

        t, unit = utils.time_writer(time() - t0)
        print(f"\nRuntime: {round(t, 2)} {unit}\n")

    def label_graph(self, graph):
        """
        Iterates over nodes in "graph" and stores the label in the
        predicted segmentation mask (i.e. "self.label_mask") which coincides
        with each node as a node-level attribute called "label".

        Parameters
        ----------
        graph : networkx.Graph
            Graph that represents a neuron from the ground truth.

        Returns
        -------
        graph : networkx.Graph
            Updated graph with node-level attributes called "label".

        """
        labeled_graph = nx.Graph(graph)
        id_to_label_nodes = dict()
        with ThreadPoolExecutor() as executor:
            # Assign threads
            threads = []
            for i in labeled_graph.nodes:
                img_coord = gutils.get_coord(labeled_graph, i)
                threads.append(executor.submit(self.get_label, img_coord, i))

            # Store results
            for thread in as_completed(threads):
                i, label = thread.result()
                labeled_graph.nodes[i].update({"label": label})
                if label in id_to_label_nodes.keys():
                    id_to_label_nodes[label].add(i)
                else:
                    id_to_label_nodes[label] = set([i])
        return labeled_graph, id_to_label_nodes

    def get_label(self, img_coord, return_node=False):
        """
        Gets label of voxel at "img_coord".

        Parameters
        ----------
        img_coord : numpy.ndarray
            Image coordinate of voxel to be read.

        Returns
        -------
        int
           Label of voxel at "img_coord".

        """
        # Read label
        if self.in_black_hole(img_coord):
            label = -1
        else:
            label = self.read_label(img_coord)

        # Adjust label
        label = self.equivalent_label(label)
        label = self.validate(label)
        if return_node:
            return return_node, label
        else:
            return label

    def read_label(self, coord):
        """
        Gets label at image coordinates "xyz".

        Parameters
        ----------
        coord : tuple[int]
            Coordinates that indexes into "self.label_mask".

        Returns
        -------
        int
           Label at image coordinates "xyz".

        """
        # Read image label
        if type(self.label_mask) == ts.TensorStore:
            return int(self.label_mask[coord].read().result())
        else:
            return self.label_mask[coord]

    def equivalent_label(self, label):
        """
        Gets the equivalence class label corresponding to "label".

        Parameters
        ----------
        label : int
            Label to be checked.

        Returns
        -------
        label
            Equivalence class label.

        """
        if self.equiv_labels_map:
            if label in self.equiv_labels_map.keys():
                return self.equiv_labels_map[label]
            else:
                return 0
        else:
            return label

    def validate(self, label):
        """
        Validates label by checking whether it is contained in
        "self.valid_labels".

        Parameters
        ----------
        label : int
            Label to be validated.

        Returns
        -------
        label : int
            There are two possibilities: (1) original label if either "label"
            is contained in "self.valid_labels" or "self.valid_labels" is
            None, or (2) 0 if "label" is not contained in self.valid_labels.

        """
        if self.valid_labels:
            return 0 if label not in self.valid_labels.keys() else label
        else:
            return label

    def init_xyz_to_id_node(self):
        self.xyz_to_id_node = dict()
        for id, graph in self.graphs.items():
            for i in graph.nodes:
                xyz = tuple(graph.nodes[i]["xyz"])
                if xyz in self.xyz_to_id_node.keys():
                    self.xyz_to_id_node[xyz][id] = i
                else:
                    self.xyz_to_id_node[xyz] = {id: i}

    def get_pred_coords(self, label):
        if self.valid_labels:
            if label in self.valid_labels.keys():
                return self.valid_labels[label]
        return []

    # -- Final Constructor Routines --
    def init_kdtree(self):
        """
        Builds a KD-Tree from the xyz coordinates from all nodes across all
        graphs contained in "self.graphs".

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        xyz_list = []
        for _, graph in self.graphs.items():
            for i in graph.nodes:
                xyz_list.append(graph.nodes[i]["xyz"])
        self.graphs_kdtree = KDTree(xyz_list)

    def get_projection(self, xyz):
        """
        Gets the xyz coordinates of the nearest neighbor of "xyz".

        Parameters
        ----------
        xyz : tuple
            xyz coordinate to be queried.

        Returns
        -------
        tuple
            xyz coordinate of the nearest neighbor of "xyz".
        d : float
            Projection distance.

        """
        d, idx = self.graphs_kdtree.query(xyz, k=1)
        return tuple(self.graphs_kdtree.data[idx]), d

    def init_black_holes(self, black_holes):
        if black_holes:
            black_holes_xyz = [bh_dict["xyz"] for bh_dict in black_holes]
            black_holes_id = [bh_dict["swc_id"] for bh_dict in black_holes]
            self.black_holes = KDTree(black_holes_xyz)
            self.black_hole_labels = set(black_holes_id)
        else:
            self.black_holes = None
            self.black_hole_labels = set()

    def in_black_hole(self, xyz):
        if self.black_holes:
            radius = self.black_hole_radius
            pts = self.black_holes.query_ball_point(xyz, radius)
            return True if len(pts) > 0 else False
        else:
            return False

    def rm_spurious_intersections(self):
        # Compute intersections
        for label in self.get_all_labels():
            hit_ids = merge_detection.intersections(
                self.get_projection,
                self.get_pred_coords(label),
                self.xyz_to_id_node,
                CLOSE_DIST_THRESHOLD,
            )

            # Remove spurious intersections
            for id in self.graphs.keys():
                if id in hit_ids.keys():
                    if len(hit_ids[id]) < INTERSECTION_THRESHOLD:
                        self.zero_nodes(id, label)
                elif label in self.id_to_label_nodes[id].keys():
                    n_hits = len(self.id_to_label_nodes[id][label])
                    if n_hits < INTERSECTION_THRESHOLD:
                        self.zero_nodes(id, label)

    # -- Evaluation --
    def compute_metrics(self):
        """
        Computes skeleton-based metrics.

        Parameters
        ----------
        None

        Returns
        -------
        ...

        """
        # Split evaluation
        print("Detecting Splits...")
        self.saved_site_cnt = 0
        self.detect_splits()
        self.quantify_splits()

        # Merge evaluation
        print("Detecting Merges...")
        self.saved_site_cnt = 0
        self.detect_merges()
        self.quantify_merges()

        # Compute metrics
        full_results, avg_results = self.compile_results()
        return full_results, avg_results

    def get_all_labels(self):
        labels = set()
        for id in self.graphs.keys():
            labels = labels.union(self.get_labels(id))
        labels.discard(0)
        return labels

    def get_labels(self, id):
        """
        Gets the predicted label ids that intersect with the target graph
        corresponding to "id".

        Parameters
        ----------
        id : str

        """
        return set(self.id_to_label_nodes[id].keys())

    def zero_nodes(self, id, label):
        """
        Zeros out nodes in "self.labeled_graph[id]" in the sense that
        nodes with "label" are updated to zero.

        Parameters
        ----------
        id : str
            ID of ground truth graph to be updated.
        label : int
            Label that identifies which nodes to have their label updated to
            zero.

        Returns
        -------
        None

        """
        if label in self.id_to_label_nodes[id].keys():
            for i in self.id_to_label_nodes[id][label]:
                self.labeled_graphs[id].nodes[i]["label"] = 0

                xyz = tuple(self.graphs[id].nodes[i]["xyz"])
                self.xyz_to_id_node[xyz][id] = 0
            del self.id_to_label_nodes[id][label]

    def detect_splits(self):
        """
        Detects splits in the predicted segmentation, then deletes node and
        edges in "self.labeled_graphs" that correspond to a split.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        t0 = time()
        for cnt, (id, graph) in enumerate(self.graphs.items()):
            # Detection
            utils.progress_bar(cnt + 1, len(self.graphs))
            labeled_graph = split_detection.run(graph, self.labeled_graphs[id])

            # Update predicted graph
            labeled_graph = gutils.delete_nodes(labeled_graph, 0)
            self.labeled_graphs[id] = gutils.delete_nodes(labeled_graph, -1)
            self.id_to_label_nodes[id] = gutils.store_labels(labeled_graph)

        # Report runtime
        t, unit = utils.time_writer(time() - t0)
        print(f"\nRuntime: {round(t, 2)} {unit}\n")

    def quantify_splits(self):
        """
        Counts the number of splits, number of omit edges, and percent of omit
        edges for each graph in "self.labeled_graphs".

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.split_cnts = dict()
        self.omit_cnts = dict()
        self.omit_percents = dict()
        for id in self.graphs.keys():
            n_pred_edges = self.labeled_graphs[id].number_of_edges()
            n_target_edges = self.graphs[id].number_of_edges()

            self.split_cnts[id] = gutils.count_splits(self.labeled_graphs[id])
            self.omit_cnts[id] = n_target_edges - n_pred_edges
            self.omit_percents[id] = 1 - n_pred_edges / n_target_edges

    def detect_merges(self):
        """
        Detects merges in the predicted segmentation, then deletes node and
        edges in "self.labeled_graphs" that correspond to a merge.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # Initilizations
        self.merge_cnts = self.init_counter()
        self.merged_cnts = self.init_counter()
        self.merged_percents = self.init_counter()

        # Check potential merge sites
        t0 = time()
        with ProcessPoolExecutor() as executor:
            processes = []
            for ids, label in self.detect_potential_merges():
                id_1, id_2 = tuple(ids)
                continue
                processes.append(
                    executor.submit(
                        merge_detection.localize,
                        self.graphs[id_1],
                        self.graphs[id_2],
                        self.id_to_label_nodes[id_1][label],
                        self.id_to_label_nodes[id_2][label],
                        MERGE_DIST_THRESHOLD,
                        (ids, label),
                    )
                )

            # Compile results
            cnt = 1
            chunk_size = int(len(processes) * 0.02)
            detected_merges = set()
            for i, process in enumerate(as_completed(processes)):
                # Check site
                merge_id, site, d = process.result()
                if d < MERGE_DIST_THRESHOLD:
                    # print(merge_id, d)
                    detected_merges.add(merge_id)
                    if self.save:
                        self.save_swc(site[0], site[1], "merge")

                # Report process
                if i > cnt * chunk_size:
                    utils.progress_bar(i + 1, len(processes))
                    cnt += 1

        # Update graph
        for (id_1, id_2), label in detected_merges:
            self.process_merge(id_1, label)
            self.process_merge(id_2, label)

        # Report Runtime
        t, unit = utils.time_writer(time() - t0)
        print(f"\nRuntime: {round(t, 2)} {unit}\n")

    def detect_potential_merges(self):
        """
        Detects merges between ground truth graphs which are considered to be
        potential merge sites.

        Parameters
        ----------
        None

        Returns
        -------
        set
            Set of tuples containing a tuple of graph ids and common label
            between the graphs.

        """
        return merge_detection.detect_potentials(
            self.labeled_graphs, self.get_labels
        )

    def near_bdd(self, xyz):
        """
        Determines whether "xyz" is near the boundary of the image.

        Parameters
        ----------
        xyz : numpy.ndarray
            xyz coordinate to be checked

        Returns
        -------
        near_bdd_bool : bool
            Indication of whether "xyz" is near the boundary of the image.

        """
        near_bdd_bool = False
        if self.ignore_boundary_mistakes:
            mask_shape = self.label_mask.shape
            above = [xyz[i] >= mask_shape[i] - 32 for i in range(3)]
            below = [xyz[i] < 32 for i in range(3)]
            near_bdd_bool = True if any(above) or any(below) else False
        return near_bdd_bool

    def init_counter(self):
        """
        Initializes a dictionary that is used to count the number of merge
        type mistakes for each pred_graph.

        Parameters
        ----------
        None

        Returns
        -------
        dict
            Dictionary used to count number of merge type mistakes.

        """
        counter = dict()
        for label in self.labeled_graphs.keys():
            counter[label] = 0
        return counter

    def init_tracker(self):
        tracker = dict()
        for label in self.labeled_graphs.keys():
            tracker[label] = set()
        return tracker

    def process_merge(self, id, label):
        """
        Once a merge has been detected that corresponds to "id", every
        node in "self.labeled_graph[id]" with that "label" is
        deleted.

        Parameters
        ----------
        id : str
            Key associated with the labeled_graph to be searched.
        label : int
            Label in prediction that is assocatied with a merge.

        Returns
        -------
        None

        """
        graph = self.labeled_graphs[id].copy()
        graph, merged_cnt = gutils.delete_nodes(graph, label, return_cnt=True)
        self.labeled_graphs[id] = graph
        self.merged_cnts[id] += merged_cnt
        self.merge_cnts[id] += 1

    def quantify_merges(self):
        """
        Computes the percentage of merged edges for each labeled_graph.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.merged_percents = dict()
        for id in self.graphs.keys():
            n_edges = self.graphs[id].number_of_edges()
            percent = self.merged_cnts[id] / n_edges
            self.merged_percents[id] = percent

    def compile_results(self):
        """
        Compiles a dictionary containing the metrics computed by this module.

        Parameters
        ----------
        None

        Returns
        -------
        full_results : dict
            Dictionary where the keys are ids and the values are the
            result of computing each metric for the corresponding graphs.
        avg_result : dict
            Dictionary where the keys are names of metrics computed by this
            module and values are the averaged result over all ids.

        """
        # Compute remaining metrics
        self.compute_edge_accuracy()
        self.compute_erl()

        # Summarize results
        ids, results = self.generate_full_results()
        avg_results = self.generate_avg_results()

        # Reformat full results
        full_results = dict()
        for i, id in enumerate(ids):
            full_results[id] = dict(
                [(key, results[key][i]) for key in results.keys()]
            )

        return full_results, avg_results

    def generate_full_results(self):
        """
        Generates a report by creating a list of the results for each metric.
        Each item in this list corresponds to a graph in labeled_graphs and
        this list is ordered with respect to "ids".

        Parameters
        ----------
        None

        Results
        -------
        grids : list[str]
            Specifies the ordering of results for each value in "stats".
        stats : dict
            Dictionary where the keys are metrics and values are the result of
            computing that metric for each graph in labeled_graphs.

        """
        ids = list(self.labeled_graphs.keys())
        ids.sort()
        stats = {
            "# splits": generate_result(ids, self.split_cnts),
            "# merges": generate_result(ids, self.merge_cnts),
            "% omit": generate_result(ids, self.omit_percents),
            "% merged": generate_result(ids, self.merged_percents),
            "edge accuracy": generate_result(ids, self.edge_accuracy),
            "erl": generate_result(ids, self.erl),
            "normalized erl": generate_result(ids, self.normalized_erl),
        }
        return ids, stats

    def generate_avg_results(self):
        avg_stats = {
            "# splits": self.avg_result(self.split_cnts),
            "# merges": self.avg_result(self.merge_cnts) / 2,
            "% omit": self.avg_result(self.omit_percents),
            "% merged": self.avg_result(self.merged_percents),
            "edge accuracy": self.avg_result(self.edge_accuracy),
            "erl": self.avg_result(self.erl),
            "normalized erl": self.avg_result(self.normalized_erl),
        }
        return avg_stats

    def avg_result(self, stats):
        result = []
        wgts = []
        for id, wgt in self.wgts.items():
            if self.omit_percents[id] < 1:
                result.append(stats[id])
                wgts.append(wgt)
        return np.average(result, weights=wgts)

    def compute_edge_accuracy(self):
        """
        Computes the edge accuracy of each self.labeled_graph.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.edge_accuracy = dict()
        for id in self.graphs.keys():
            omit_percent = self.omit_percents[id]
            merged_percent = self.merged_percents[id]
            self.edge_accuracy[id] = 1 - omit_percent - merged_percent

    def compute_erl(self):
        """
        Computes the expected run length (ERL) of each labeled_graph.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.erl = dict()
        self.normalized_erl = dict()
        self.wgts = dict()
        total_path_length = 0
        for id in self.graphs.keys():
            labeled_graph = self.labeled_graphs[id]
            path_length = gutils.compute_path_length(self.graphs[id])
            run_lengths = gutils.compute_run_lengths(labeled_graph)
            wgt = run_lengths / max(np.sum(run_lengths), 1)

            self.erl[id] = np.sum(wgt * run_lengths)
            self.normalized_erl[id] = self.erl[id] / path_length

            self.wgts[id] = path_length
            total_path_length += path_length

        for id in self.graphs.keys():
            self.wgts[id] = self.wgts[id] / total_path_length

    def list_metrics(self):
        """
        Lists metrics that are computed by this module.

        Parameters
        ----------
        None

        Returns
        -------
        metrics : list[str]
            List of metrics computed by this module.

        """
        metrics = [
            "# splits",
            "# merges",
            "% omit",
            "% merged",
            "edge accuracy",
            "erl",
            "normalized erl",
        ]
        return metrics

    def save_swc(self, xyz_1, xyz_2, mistake_type):
        self.saved_site_cnt += 1
        xyz_1 = utils.to_world(xyz_1, self.anisotropy)
        xyz_2 = utils.to_world(xyz_2, self.anisotropy)
        color = "0.0 1.0 0.0" if mistake_type == "split" else "0.0 0.0 1.0"
        path = f"{self.output_dir}/{mistake_type}-{self.saved_site_cnt}.swc"
        save(path, xyz_1, xyz_2, color=color)


# -- utils --
def generate_result(ids, stats):
    """
    Reorders items in "stats" with respect to the order defined by "ids".

    Parameters
    ----------
    ids : list[str]
        List of all "ids" of graphs in "self.labeled_graphs".
    stats : dict
        Dictionary where the keys are "ids" and values are the result
        of computing some metrics.

    Returns
    -------
    list
        Reorded items in "stats" with respect to the order defined by
        "ids".

    """
    return [stats[id] for id in ids]



# -*- coding: utf-8 -*-
"""
Created on Wed June 5 16:00:00 2023

@author: Anna Grim
@email: anna.grim@alleninstitute.org

"""

from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from zipfile import ZipFile

import networkx as nx
import numpy as np
from google.cloud import storage

from segmentation_skeleton_metrics import utils


def parse(swc_paths, min_size, anisotropy):
    """
    Reads swc files and extracts the xyz coordinates.

    Paramters
    ---------
    swc_paths : list or dict
        If swc files are on local machine, list of paths to swc files where
        each file corresponds to a neuron in the prediction. If swc files are
        on cloud, then dict with keys "bucket_name" and "path".
    min_size : int
        Threshold on the number of nodes contained in an swc file. Only swc
        files with more than "min_size" nodes are stored in "valid_labels".
    anisotropy : list[float]
        Image to World scaling factors applied to xyz coordinates to account
        for anisotropy of the microscope.

    Returns
    -------
    dict
        Dictionary where each item is an swc_id and an array of the xyz
        coordinates read from cooresponding swc file.

    """
    if type(swc_paths) == list:
        return parse_local_paths(swc_paths, min_size, anisotropy)
    elif type(swc_paths) == dict:
        return parse_cloud_paths(swc_paths, min_size, anisotropy)
    else:
        return None


def parse_local_paths(swc_paths, min_size, anisotropy):
    """
    Reads swc files from local machine and extracts the xyz coordinates.

    Paramters
    ---------
    swc_paths : list or dict
        If swc files are on local machine, list of paths to swc files where
        each file corresponds to a neuron in the prediction. If swc files are
        on cloud, then dict with keys "bucket_name" and "path".
    min_size : int
        Threshold on the number of nodes contained in an swc file. Only swc
        files with more than "min_size" nodes are stored in "valid_labels".
    anisotropy : list[float]
        Image to World scaling factors applied to xyz coordinates to account
        for anisotropy of the microscope.

    Returns
    -------
    valid_labels : dict
        Dictionary where each item is an swc_id and an array of the xyz
        coordinates read from cooresponding swc file.

    """
    valid_labels = dict()
    for path in swc_paths:
        contents = read_from_local(path)
        if len(contents) > min_size:
            id = int(utils.get_id(path))
            valid_labels[id] = get_coords(contents, anisotropy)
    return valid_labels


def parse_cloud_paths(cloud_dict, min_size, anisotropy):
    """
    Reads swc files from a GCS bucket and extracts the xyz coordinates.

    Parameters
    ----------
    cloud_dict : dict
        Dictionary where keys are "bucket_name" and "path".
    min_size : int
        Threshold on the number of nodes contained in an swc file. Only swc
        files with more than "min_size" nodes are stored in "valid_labels".
    anisotropy : list[float]
        Image to World scaling factors applied to xyz coordinates to account
        for anisotropy of the microscope.

    Returns
    -------
    valid_labels : dict
        Dictionary where each item is an swc_id and an array of the xyz
        coordinates read from cooresponding swc file.

    """
    # Initializations
    bucket = storage.Client().bucket(cloud_dict["bucket_name"])
    zip_paths = utils.list_gcs_filenames(bucket, cloud_dict["path"], ".zip")
    chunk_size = int(len(zip_paths) * 0.02)

    # Parse
    cnt = 1
    valid_labels = dict()
    print("Downloading predicted swc files from cloud...")
    print("# zip files:", len(zip_paths))
    for i, path in enumerate(zip_paths):
        valid_labels.update(download(bucket, path, min_size, anisotropy))
        if i > cnt * chunk_size:
            utils.progress_bar(i + 1, len(zip_paths))
            cnt += 1
        break

    # Report Results
    print("\n#Valid Labels:", len(valid_labels))
    print("")
    return valid_labels


def download(bucket, zip_path, min_size, anisotropy):
    """
    Downloads the contents from each swc file contained in the zip file at
    "zip_path".

    Parameters
    ----------
    bucket : str
        Name of GCS bucket containing swc files to be read.
    zip_path : str
        Path to zip file contained in GCS bucket.
    min_size : int
        Threshold on the number of nodes contained in an swc file. Only swc
        files with more than "min_size" nodes are stored in "valid_labels".
    anisotropy : list[float]
        Image to World scaling factors applied to xyz coordinates to account
        for anisotropy of the microscope.

    Returns
    -------
    valid_labels : dict
        Dictionary where each item is an swc_id and an array of the xyz
        coordinates read from cooresponding swc file.

    """
    zip_content = bucket.blob(zip_path).download_as_bytes()
    with ZipFile(BytesIO(zip_content)) as zip_file:
        with ThreadPoolExecutor() as executor:
            # Assign threads
            threads = []
            for path in utils.list_files_in_gcs_zip(zip_content):
                threads.append(
                    executor.submit(
                        parse_gcs_zip, zip_file, path, min_size, anisotropy
                    )
                )

            # Process results
            valid_labels = dict()
            for thread in as_completed(threads):
                valid_labels.update(thread.result())
    return valid_labels


def parse_gcs_zip(zip_file, path, min_size, anisotropy):
    """
    Reads swc file stored at "path" which points to a file in a GCS bucket.

    Parameters
    ----------
    zip_file : ZipFile
        Zip file containing swc file to be read.
    path : str
        Path to swc file to be read.
    min_size : int
        Threshold on the number of nodes contained in an swc file. Only swc
        files with more than "min_size" nodes are stored in "valid_labels".
    anisotropy : list[float]
        Image to World scaling factors applied to xyz coordinates to account
        for anisotropy of the microscope.

    Returns
    -------
    list
        List such that each entry is a line from the swc file.

    """
    contents = read_from_cloud(zip_file, path)
    if len(contents) > min_size:
        swc_id = int(utils.get_id(path))
        return {swc_id: get_coords(contents, anisotropy)}
    else:
        return dict()


def read_from_local(path):
    """
    Reads swc file stored at "path" on local machine.

    Parameters
    ----------
    path : str
        Path to swc file to be read.

    Returns
    -------
    list
        List such that each entry is a line from the swc file.

    """
    with open(path, "r") as file:
        return file.readlines()


def read_from_cloud(zip_file, path):
    """
    Reads the content of an swc file from a zip file in a GCS bucket.

    """
    with zip_file.open(path) as text_file:
        return text_file.read().decode("utf-8").splitlines()


def get_coords(contents, anisotropy):
    """
    Gets the xyz coords from the swc file at "path".

    Parameters
    ----------
    path : str
        Path to swc file to be parsed.
    anisotropy : list[float]
        Image to World scaling factors applied to xyz coordinates to account
        for anisotropy of the microscope.

    Returns
    -------
    numpy.ndarray
        xyz coords from an swc file.

    """
    coords_list = []
    offset = [0, 0, 0]
    for line in contents:
        if line.startswith("# OFFSET"):
            parts = line.split()
            offset = read_xyz(parts[2:5])
        if not line.startswith("#"):
            parts = line.split()
            coord = read_xyz(parts[2:5], anisotropy=anisotropy, offset=offset)
            coords_list.append(coord)
    return np.array(coords_list)


def read_xyz(xyz, anisotropy=[1.0, 1.0, 1.0], offset=[0, 0, 0]):
    """
    Reads the (x,y,z) coordinates from an swc file, then reverses and scales
    them.

    Parameters
    ----------
    coord : str
        xyz coordinate.
    anisotropy : list[float], optional
        Image to real-world coordinates scaling factors applied to "xyz". The
        default is [1.0, 1.0, 1.0].
    offset : list[int], optional
        Offset of (x, y, z) coordinates in swc file. The default is [0, 0, 0].

    Returns
    -------
    tuple
        The (x,y,z) coordinates of an entry from an swc file in real-world
        coordinates.

    """
    xyz = [float(xyz[i]) + offset[i] for i in range(3)]
    return np.array([xyz[i] / anisotropy[i] for i in range(3)], dtype=int)


def save(path, xyz_1, xyz_2, color=None):
    """
    Writes an swc file.

    Parameters
    ----------
    path : str
        Path on local machine that swc file will be written to.
    entry_list : list[list]
        List of entries that will be written to an swc file.
    color : str, optional
        Color of nodes. The default is None.

    Returns
    -------
    None.

    """
    with open(path, "w") as f:
        # Preamble
        if color is not None:
            f.write("# COLOR " + color)
        else:
            f.write("# id, type, z, y, x, r, pid")
        f.write("\n")

        # Entries
        f.write(make_entry(1, -1, xyz_1))
        f.write("\n")
        f.write(make_entry(2, 1, xyz_2))


def make_entry(node_id, parent_id, xyz):
    """
    Makes an entry to be written in an swc file.

    Parameters
    ----------
    graph : networkx.Graph
        Graph that "node_id" and "parent_id" belong to.
    node_id : int
        Node that entry corresponds to.
    parent_id : int
         Parent of node "node_id".
    xyz : ...
        xyz coordinate of "node_id".

    Returns
    -------
    entry : str
        Entry to be written in an swc file.

    """
    x, y, z = tuple(xyz)
    entry = f"{node_id} 2 {x} {y} {z} 8 {parent_id}"
    return entry


def to_graph(path, anisotropy=[1.0, 1.0, 1.0]):
    """
    Reads an swc file and builds an undirected graph from it.

    Parameters
    ----------
    path : str
        Path to swc file to be read.
    anisotropy : list[float], optional
        Image to real-world coordinates scaling factors for (x, y, z) that is
        applied to swc files. The default is [1.0, 1.0, 1.0].

    Returns
    -------
    networkx.Graph
        Graph built from an swc file.

    """
    graph = nx.Graph(swc_id=utils.get_id(path))
    offset = [0, 0, 0]
    for line in read_from_local(path):
        if line.startswith("# OFFSET"):
            parts = line.split()
            offset = read_xyz(parts[2:5])
        if not line.startswith("#"):
            parts = line.split()
            child = int(parts[0])
            parent = int(parts[-1])
            xyz = read_xyz(parts[2:5], anisotropy=anisotropy, offset=offset)
            graph.add_node(child, xyz=xyz)
            if parent != -1:
                graph.add_edge(parent, child)
    return graph
